from logging import getLogger

import numpy as np

logger = getLogger(__name__)

def wer_tabular(hyp: list[str], ref: list[str]):
    table = np.zeros((len(hyp) + 1, len(ref) + 1), dtype=np.uint8)
    table[:, 0] = np.arange(len(hyp) + 1)
    table[0, :] = np.arange(len(ref) + 1)

    for i in range(1, len(hyp) + 1):
        for j in range(1, len(ref) + 1):
            table[i, j] = min(
                table[i - 1, j] + 1,
                table[i, j - 1] + 1,
                table[i - 1, j - 1] + (0 if hyp[i - 1] == ref[j - 1] else 1)
            )

    return table[-1, -1]

def calculate_wer(hyp: list[str], ref: list[str]):
    word_error_count = wer_tabular(hyp, ref)
    return word_error_count / len(ref)


def num_tokens_from_messages(messages, model="gpt-3.5-turbo-0613"):
  """Returns the number of tokens used by a list of messages."""
  try:
      encoding = tiktoken.encoding_for_model(model)
  except KeyError:
      encoding = tiktoken.get_encoding("cl100k_base")
  if model == "gpt-3.5-turbo-0613":  # note: future models may deviate from this
      num_tokens = 0
      for message in messages:
          num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
          for key, value in message.items():
              num_tokens += len(encoding.encode(value))
              if key == "name":  # if there's a name, the role is omitted
                  num_tokens += -1  # role is always required and always 1 token
      num_tokens += 2  # every reply is primed with <im_start>assistant
      return num_tokens
  else:
      raise NotImplementedError(f"""num_tokens_from_messages() is not presently implemented for model {model}.
  See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens.""")


import re
import pathlib
import copy
from enum import Enum
from dataclasses import dataclass
from typing import IO, Optional, BinaryIO
from collections import namedtuple
from logging import getLogger

import pydantic
from pdfminer.pdfdocument import PDFDocument, PDFNoOutlines
from pdfminer.pdfparser import PDFParser
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.converter import PDFPageAggregator
from pdfminer.layout import LTPage, LTChar, LTAnno, LAParams, LTTextBox, LTTextLine, LTContainer
from pdfminer.pdfpage import PDFPage

from patent_chart.google_patents import parse_google_patent
from patent_chart.data_structures import PatentTextLine

logger = getLogger(__name__)

StrBytePath = str | IO | pathlib.Path

claim_number_regex = re.compile(r'\s*\d{1,2}\.')

class TokenType(Enum):
    DELIMITER_START_CLAIM = 'DELIMITER_START_CLAIM'
    # DELIMITER_END_CLAIM = 'DELIMITER_END_CLAIM'
    DELIMITER_END_PREAMBLE = 'DELIMITER_END_PREAMBLE'
    DELIMITER_END_ELEMENT = 'DELIMITER_END_ELEMENT'
    TEXT = 'TEXT'
    WHITESPACE = 'WHITESPACE'
    DEPENDENCY = 'DEPENDENCY'

@dataclass
class Token:
    type: TokenType
    lexeme: str

class ClaimsLexer(Enum):
    # TODO: move start claims section into here instead of as a pre-processing step. That way we can just run this on the whole patent rather than pre-processing the claims first. If we do that, also uncomment the end claims section delimiter.
    # DELIMITER_END_CLAIMS_SECTION = re.compile(r'k k k k k')
    DELIMITER_START_CLAIM = re.compile(r'[1-9][0-9]*\s*\.')
    # TODO: handle multiple dependency, which would include language like:
    #   claim 3 or 4
    #   claim 1 or claim 2
    #   claim 1, 7, 12, or 15
    #   claim 1, claim 7, claim 12, or claim 15
    #   any of the preceding claims
    DEPENDENCY = re.compile(r'[Cc]laim [1-9][0-9]*')
    DELIMITER_END_PREAMBLE = re.compile(r':')
    DELIMITER_END_ELEMENT = re.compile(r';')
    # DELIMITER_END_CLAIM = re.compile(r'\.')
    # TODO: add transition type:
    #   comprising
    #   consisting of
    #   consisting essentially of
    #   etc...
    # TEXT = re.compile(r'[\w,\'\"!?\(\)\-—/]+')
    TEXT = re.compile(r'[\w,\'\"!?\(\)\-—/\.]+')  # Text including period
    WHITESPACE = re.compile(r'\s+')

    def __repr__(self):
        return self.name
    
class LexicalError(Exception):
    pass

class ParseError(Exception):
    pass

def lex_claims(claim_text):
    tokens = []
    while claim_text:
        token_matched = False
        for token_type in ClaimsLexer:
            match = token_type.value.match(claim_text)
            if match:
                token_matched = True
                tokens.append(Token(getattr(TokenType, token_type.name), lexeme=match.group()))
                claim_text = claim_text[match.end():]
                break
        if not token_matched:
            raise LexicalError(f'No token matched for claim text: {claim_text[:10]}')
    return tokens

# Base recursive descent parser
class RecursiveDescentParserBase:
    def __init__(self, tokens):
        self.tokens = tokens
        self.current_token_index = 0
        
    def peek(self):
        return self.tokens[self.current_token_index]
    
    def is_at_end(self):
        return self.current_token_index == len(self.tokens)
    
    def previous(self):
        return self.tokens[self.current_token_index - 1]

    def advance(self):
        self.current_token_index += 1
        return self.previous()

    def reverse(self):
        self.current_token_index -= 1
        return self.previous()

    def check(self, type):
        if self.is_at_end():
            return False
        return type == self.peek().type

    def match(self, *token_types):
        for token_type in token_types:
            if self.check(token_type):
                self.advance()
                return True
        return False
    
    def consume(self, token_type):
        if self.match(token_type):
            return self.previous()
        else:
            raise Exception(f'Expected {token_type.name}')

    def parse(self):
        pass

# Define dataclasses for concrete syntax tree for claims parser
# NOTE: order of definition matters for pydantic validation
@dataclass
class ClaimText:
    text: Token

@dataclass
class ClaimDelimiterEndPreamble:
    delimiter_end_preamble: Token

@dataclass
class ClaimDelimiterEndElement:
    delimiter_end_element: Token

@dataclass
class ClaimWhitespace:
    whitespace: Token

@dataclass
class ClaimDependency:
    dependency: Token

# @dataclass
# class Claim:
#     delimiter_start_claim: Token
#     claim_content: list[ClaimText | ClaimDelimiterEndPreamble |ClaimDelimiterEndElement | ClaimWhitespace | ClaimDependency]
#     delimiter_end_claim: Token

@dataclass
class GoogleClaim:
    claim_number: int
    claim_elements: list[str]

@dataclass
class Claim:
    delimiter_start_claim: Token
    claim_content: list[ClaimText | ClaimDelimiterEndPreamble |ClaimDelimiterEndElement | ClaimWhitespace | ClaimDependency]

@dataclass
class Claims:
    claims: list[Claim]

class ClaimsParser(RecursiveDescentParserBase):
    def __init__(self, tokens):
        self.tokens = tokens
        self.current_token_index = 0

    def claim_content(self):
        if self.match(TokenType.TEXT):
            text = self.previous()
            return ClaimText(text=text)
        elif self.match(TokenType.WHITESPACE):
            whitespace = self.previous()
            return ClaimWhitespace(whitespace=whitespace)
        elif self.match(TokenType.DEPENDENCY):
            dependency = self.previous()
            return ClaimDependency(dependency=dependency)
        elif self.match(TokenType.DELIMITER_END_PREAMBLE):
            delimiter_end_preamble = self.previous()
            return ClaimDelimiterEndPreamble(delimiter_end_preamble=delimiter_end_preamble)
        elif self.match(TokenType.DELIMITER_END_ELEMENT):
            delimiter_end_element = self.previous()
            return ClaimDelimiterEndElement(delimiter_end_element=delimiter_end_element)
        else:
            raise Exception('Expected claim text')
            
    def claim(self):
        if self.match(TokenType.DELIMITER_START_CLAIM):
            delimiter_start_claim = self.previous()
            claim_contents = []
            # while not self.match(TokenType.DELIMITER_END_CLAIM):
            #     claim_contents.append(self.claim_content())
            # delimiter_end_claim = self.previous()
            # return Claim(delimiter_start_claim=delimiter_start_claim, claim_content=claim_contents, delimiter_end_claim=delimiter_end_claim)
            while not self.check(TokenType.DELIMITER_START_CLAIM) and not self.is_at_end():
                claim_contents.append(self.claim_content())
            return Claim(delimiter_start_claim=delimiter_start_claim, claim_content=claim_contents)
        else:
            raise Exception('Expected claim start delimiter')
    
    def claims(self):
        claims = []
        while not self.is_at_end():
            # There can be whitespace and text between claims. So we need to chew it up
            # before parsing the next claim.
            while self.match(TokenType.WHITESPACE, TokenType.TEXT):
                continue
            if not self.is_at_end():
                claim = self.claim()
                claims.append(claim)
        return Claims(claims=claims)

    def parse(self):
        return self.claims()

def serialize_claim(claim):
    serialized = ''
    serialized += claim.delimiter_start_claim.lexeme
    for claim_content in claim.claim_content:
        if isinstance(claim_content, ClaimText):
            serialized += claim_content.text.lexeme
        elif isinstance(claim_content, ClaimWhitespace):
            serialized += claim_content.whitespace.lexeme
        elif isinstance(claim_content, ClaimDependency):
            serialized += claim_content.dependency.lexeme
        elif isinstance(claim_content, ClaimDelimiterEndPreamble):
            serialized += claim_content.delimiter_end_preamble.lexeme
        elif isinstance(claim_content, ClaimDelimiterEndElement):
            serialized += claim_content.delimiter_end_element.lexeme
    # serialized += claim.delimiter_end_claim.lexeme
    return serialized

def serialize_claim_elements(claim):
    serialized_claim_elements = []
    # Iterate through claim content, incrementally constructing a claim element. Once a delimiter end preamble or delimiter end element is encountered, add the claim element to the list of claim elements and start a new claim element.
    claim_element = claim.delimiter_start_claim.lexeme
    for claim_content in claim.claim_content:
        if isinstance(claim_content, ClaimText):
            claim_element += claim_content.text.lexeme
        elif isinstance(claim_content, ClaimWhitespace):
            claim_element += claim_content.whitespace.lexeme
        elif isinstance(claim_content, ClaimDependency):
            claim_element += claim_content.dependency.lexeme
        elif isinstance(claim_content, ClaimDelimiterEndPreamble):
            claim_element += claim_content.delimiter_end_preamble.lexeme
            serialized_claim_elements.append(claim_element)
            claim_element = ''
        elif isinstance(claim_content, ClaimDelimiterEndElement):
            claim_element += claim_content.delimiter_end_element.lexeme
            serialized_claim_elements.append(claim_element)
            claim_element = ''
    # Add the last claim element
    # claim_element += claim.delimiter_end_claim.lexeme
    serialized_claim_elements.append(claim_element)
    # Strip whitespace from claim elements
    serialized_claim_elements = [claim_element.strip() for claim_element in serialized_claim_elements]
    return serialized_claim_elements

@dataclass
class TextLine:
    page_number: int
    x0: float
    y0: float
    x1: float
    y1: float
    text: str

@dataclass
class TextLines:
    column_0: list[TextLine]
    column_1: list[TextLine]
    noncolumnar: list[TextLine]


class PDFPageExtremaXStartAndEndAggregator(PDFPageAggregator):
    def __init__(self, rsrcmgr, pageno=1, laparams=None, bottom_margin=50):
        PDFPageAggregator.__init__(self, rsrcmgr, pageno=pageno, laparams=laparams)
        """
        Find the median of the 10 highest values for max and the median of the 10 lowest values for min.
        """
        self.lttextline_bbox_min_x0_by_page = {}
        self.lttextline_bbox_max_x1_by_page = {}
        self.bottom_margin = bottom_margin
        self.page_number = 0

    def receive_layout(self, ltpage):        
        def render(item, page_number):
            if isinstance(item, LTPage) or isinstance(item, LTTextBox):
                for child in item:
                    render(child, page_number)
            elif isinstance(item, LTTextLine):
                if item.bbox[1] <= self.bottom_margin:
                    return
                if page_number not in self.lttextline_bbox_min_x0_by_page:
                    self.lttextline_bbox_min_x0_by_page[page_number] = [item.bbox[0]]
                    self.lttextline_bbox_max_x1_by_page[page_number] = [item.bbox[2]]
                else:
                    self.lttextline_bbox_min_x0_by_page[page_number].append(item.bbox[0])
                    self.lttextline_bbox_max_x1_by_page[page_number].append(item.bbox[2])
                    
                for child in item:
                    render(child, page_number)
            return
        render(ltpage, self.page_number)
        # Reduce to median of 10 highest values for max and median of 10 lowest values for min
        if self.page_number in self.lttextline_bbox_min_x0_by_page:
            # If there were no LTTextLine on the page, then we don't have any min or max values to take the median of.
            self.lttextline_bbox_max_x1_by_page[self.page_number] = sorted(self.lttextline_bbox_max_x1_by_page[self.page_number])[-10:]
            self.lttextline_bbox_min_x0_by_page[self.page_number] = sorted(self.lttextline_bbox_min_x0_by_page[self.page_number])[:10]
            # Now take the midpoint
            self.lttextline_bbox_min_x0_by_page[self.page_number] = self.lttextline_bbox_min_x0_by_page[self.page_number][len(self.lttextline_bbox_min_x0_by_page[self.page_number]) // 2]
            self.lttextline_bbox_max_x1_by_page[self.page_number] = self.lttextline_bbox_max_x1_by_page[self.page_number][len(self.lttextline_bbox_max_x1_by_page[self.page_number]) // 2]
        self.page_number += 1
        self.result = ltpage

class PDFPageDetailedAggregator(PDFPageAggregator):
    def __init__(self, rsrcmgr, pageno=1, laparams=None, check_for_div_by_5_line_no=False, bottom_margin=50, lttextline_bbox_min_x0_by_page=None, lttextline_bbox_max_x1_by_page=None, default_to_page_width=True, pages=None):
        PDFPageAggregator.__init__(self, rsrcmgr, pageno=pageno, laparams=laparams)
        self.columns = [[],[]]
        # If parsed document is noncolumnar, then this becomes isomorphic to self.rows and self.columns should be empty.
        self.noncolumnar = []
        self.page_bboxes = []
        self.page_number = 0
        self.check_for_div_by_5_line_no = check_for_div_by_5_line_no
        # In points. Any line that is this many points or less from the bottom of the page is considered to be part of the footer and is ignored.
        self.bottom_margin = bottom_margin
        self.default_to_page_width = default_to_page_width
        self.lttextline_bbox_min_x0_by_page = lttextline_bbox_min_x0_by_page
        self.lttextline_bbox_max_x1_by_page = lttextline_bbox_max_x1_by_page

        self.text_lines = []
        self.non_text_lines = []

        self.pages = pages

    def receive_layout(self, ltpage):        
        def render(item, page_number):
            if isinstance(item, LTPage) or isinstance(item, LTTextBox):
                for child in item:
                    render(child, page_number)
            elif isinstance(item, LTTextLine):
                if item.bbox[1] <= self.bottom_margin:
                    return
                self.text_lines.append(item)
                child_str = ''
                for child in item:
                    if isinstance(child, (LTChar, LTAnno)):
                        child_str += child.get_text()
                child_str = ' '.join(child_str.split()).strip()
                if child_str:
                    row = TextLine(page_number, item.bbox[0], item.bbox[1], item.bbox[2], item.bbox[3], child_str) # bbox == (x1, y1, x2, y2)
                    x_center_of_page = ltpage.width / 2
                    # Check ltpage.width and item.bbox values are off by order of magnitude
                    off_by_order_of_magnitude = False
                    if ltpage.width / item.bbox[2] > 10 or item.bbox[2] / ltpage.width > 10:
                        off_by_order_of_magnitude = True
                    if (not self.default_to_page_width or off_by_order_of_magnitude) and self.lttextline_bbox_min_x0_by_page is not None and self.lttextline_bbox_max_x1_by_page is not None and page_number in self.lttextline_bbox_min_x0_by_page and page_number in self.lttextline_bbox_max_x1_by_page:
                        x_center_of_page = (self.lttextline_bbox_min_x0_by_page[page_number] + self.lttextline_bbox_max_x1_by_page[page_number]) / 2
                    if item.bbox[2] <= x_center_of_page:
                        self.columns[0].append(row)
                    elif item.bbox[0] >= x_center_of_page:
                        self.columns[1].append(row)
                    else:
                        if self.check_for_div_by_5_line_no:
                            # If this line is in the left column and its last one or two characters are digits that form an integer divisible by 5, then we assume it is a line number and not part of the text of the patent. Therefore we shouldn't take into account the added character width of the line number when calculating whether the line is in the left or right column. So we'll find the last character before the line number and substitute its x1 value for the item's x1 value.
                            children = list(item)
                            # check if greater overlap with left column or right column
                            # if children[0].get_text() == '5' and page_number == 6 and children[2].get_text() == '\n':
                            #     import pdb; pdb.set_trace()
                            if item.bbox[2] - x_center_of_page < x_center_of_page - item.bbox[0]:
                                # Greater overlap with left column
                                i = -1
                                while children[i].get_text().isspace():
                                    i -= 1
                                if children[i].get_text().isdigit() and int(children[i].get_text()) % 5 == 0:
                                    i -= 1
                                    if len(children) > -i and children[i].get_text().isdigit():
                                        i -= 1
                                    if len(children) > -i:
                                        # Maybe eat more whitespace
                                        while isinstance(children[i], LTAnno) or children[i].get_text().isspace():
                                            i -= 1
                                    # Maybe this is a weird short text line hanging out in the middle of the page
                                    i = max(i, -len(children))
                                    if children[i].x1 <= x_center_of_page:
                                        self.columns[0].append(row)
                                    else:
                                        self.noncolumnar.append(row)
                                else:
                                    self.noncolumnar.append(row)
                            else:
                                # TODO: fix this to mirror the greater overlap with left column branch
                                # Greater overlap with right column
                                # check if first character in line is digit divisible by 5
                                if children[0].get_text().isdigit() and int(children[0].get_text()) % 5 == 0:
                                    new_x0 = item.bbox[0] + children[0].width
                                    row.text = row.text[1:]
                                    if len(children) > 1 and children[1].get_text().isdigit():
                                        new_x0 += children[1].width
                                        row.text = row.text[1:]
                                    
                                    if new_x0 >= x_center_of_page:
                                        self.columns[1].append(row)
                                    else:
                                        self.noncolumnar.append(row)
                                else:
                                    self.noncolumnar.append(row)
                        else:
                            self.noncolumnar.append(row)
                for child in item:
                    render(child, page_number)
            return

        if self.pages is not None and self.page_number not in self.pages:
            # TODO: doesnt really help much probably because whole file still being loaded and processed before this step.
            self.page_number += 1
            return
        render(ltpage, self.page_number)
        self.page_number += 1
        self.page_bboxes.append(ltpage.bbox)
        # self.rows = sorted(self.rows, key = lambda x: (x.page_number, -x.y0, -x.x0))
        cmp_fn = lambda x: (x.page_number, -x.y0)
        self.columns[0] = sorted(self.columns[0], key=cmp_fn)
        self.columns[1] = sorted(self.columns[1], key=cmp_fn)
        self.noncolumnar = sorted(self.noncolumnar, key=cmp_fn)
        self.result = ltpage

def parse_text_lines_from_pdf_path(path: pathlib.Path | str) -> TextLines:
    with open(path, 'rb') as f:
        return parse_text_lines_from_pdf_file_obj(f)

def parse_text_lines_from_pdf_path_simple(path: pathlib.Path | str, **aggregator_params) -> TextLines:
    with open(path, 'rb') as f:
        return parse_text_lines_from_pdf_file_obj_simple(f, **aggregator_params)

def parse_text_lines_from_pdf_file_obj_simple(f: BinaryIO, **aggregator_params) -> TextLines:
    parser = PDFParser(f)
    doc = PDFDocument(parser)

    rsrcmgr = PDFResourceManager()
    # TODO: with default params, if the line number in the center margin precedes text that is too far away in the x direction, then the line number and the text get parsed as two separate lines. This typically happens when the text is indented for some reason, for example, in the claims section of patent (see Ellis patent)
    laparams = LAParams()

    device = PDFPageDetailedAggregator(rsrcmgr, laparams=laparams, **aggregator_params)
    interpreter = PDFPageInterpreter(rsrcmgr, device)

    for page in PDFPage.create_pages(doc):
        interpreter.process_page(page)
        # receive the LTPage object for this page
        device.get_result()

    return TextLines(
        column_0=device.columns[0],
        column_1=device.columns[1],
        noncolumnar=device.noncolumnar,
    )

def parse_text_lines_from_pdf_file_obj(f: BinaryIO) -> TextLines:
    parser = PDFParser(f)
    doc = PDFDocument(parser)

    rsrcmgr = PDFResourceManager()
    # TODO: with default params, if the line number in the center margin precedes text that is too far away in the x direction, then the line number and the text get parsed as two separate lines. This typically happens when the text is indented for some reason, for example, in the claims section of patent (see Ellis patent)
    laparams = LAParams()

    # First find the min x0 and max x0 per page
    device = PDFPageExtremaXStartAndEndAggregator(rsrcmgr, laparams=laparams)
    interpreter = PDFPageInterpreter(rsrcmgr, device)
    for page in PDFPage.create_pages(doc):
        interpreter.process_page(page)
        # receive the LTPage object for this page
        device.get_result()

    # Seek file back to beginning and parse again
    f.seek(0)

    device = PDFPageDetailedAggregator(rsrcmgr, laparams=laparams, check_for_div_by_5_line_no=True, lttextline_bbox_min_x0_by_page=device.lttextline_bbox_min_x0_by_page, lttextline_bbox_max_x1_by_page=device.lttextline_bbox_max_x1_by_page)
    interpreter = PDFPageInterpreter(rsrcmgr, device)

    for page in PDFPage.create_pages(doc):
        interpreter.process_page(page)
        # receive the LTPage object for this page
        device.get_result()

    return TextLines(
        column_0=device.columns[0],
        column_1=device.columns[1],
        noncolumnar=device.noncolumnar,
    )

# Succeeded by optional whitespace or zeroes
country_code_pattern = r'(?P<country_code>[A-Z]{2})\s*[O0]*'
optional_year = r'(?P<year>[1-2][0-9]{3})?'
# Negative lookahead to make sure patent number pattern doesnt match year. There should be no digits after the patent number 
patent_number_pattern = r'(?P<patent_number>[0-9]{6,8})(?!\d+)'
# Sometimes B1 can be parsed by pdfminer as Bl
kind_code_pattern = r'(?P<kind_code>[A-C]{1}[\dl]?)?'

patent_number_regex = re.compile(country_code_pattern + optional_year + patent_number_pattern + kind_code_pattern)

@dataclass
class PatentUniqueID:
    patent_number: str
    country_code: str
    kind_code: Optional[str] = None


def find_beginning_of_specification(column_0: list[PatentTextLine], column_1: list[PatentTextLine]) -> int:
    """
    Returns page number, assumes spec starts at first line of page
    """
    def beg_spec_classification_model(page_text):
        keywords = ['BACKGROUND', 'DESCRIPTION', 'SUMMARY']
        for keyword in keywords:
            if keyword in page_text:
                return True
        return False
    
    # Check column 0 
    for text_line in column_0:
        if beg_spec_classification_model(text_line.text):
            return text_line.page_num
    # Check column 1
    for text_line in column_1:
        if beg_spec_classification_model(text_line.text):
            return text_line.page_num
    return None

BeginningOfClaims = namedtuple('BeginningOfClaims', ['page_index', 'col_index', 'line_index'])
EndOfClaims = namedtuple('EndOfClaims', ['page_index', 'col_index', 'line_index'])

def find_beginning_of_claims(column_0: list[PatentTextLine], column_1: list[PatentTextLine]) -> BeginningOfClaims | None:
    """ 
    Returns page number and line number where claims start
    """
    def beg_claims_classification_model(page_text):
        # TODO: replace this with a more robust model
        key_phrases = ['What is claimed is:', 'We claim:', 'I claim:', 'The invention claimed is:', 'The invention claimed is as follows:', 'The invention claimed is:', 'claim:', 'Claims:', 'is claimed is']
        for key_phrase in key_phrases:
            if key_phrase in page_text:
                return True
        return False
    
    # Check column 0
    for text_line in column_0:
        if beg_claims_classification_model(text_line.text):
            return BeginningOfClaims(page_index=text_line.page_num, col_index=0, line_index=text_line.line_index)
        
    # Check column 1
    for text_line in column_1:
        if beg_claims_classification_model(text_line.text):
            return BeginningOfClaims(page_index=text_line.page_num, col_index=1, line_index=text_line.line_index)
    return None

class EndClaimsClassificationModel:
    def __init__(self):
        self.key_phrases = ['k k k k k', '* * * * *']
        self.key_phrases_built = ['' for _ in self.key_phrases]
        self.key_phrase_beginning_line = [None for _ in self.key_phrases]
        self.key_phrases_len = [len(''.join(key_phrase.split())) for key_phrase in self.key_phrases]
        self.key_phrases_iter = [self.iter_key_phrase(key_phrase) for key_phrase in self.key_phrases]

    def iter_key_phrase(self, key_phrase):
        for char in key_phrase:
            if char.isspace():
                continue
            else:
                yield char

    def classify(self, line: PatentTextLine) -> PatentTextLine | None:
        for key_phrase in self.key_phrases:
            if key_phrase in line.text:
                return line
        # Check if key phrase is split across lines
        for i, key_phrase_iter in enumerate(self.key_phrases_iter):
            char = next(key_phrase_iter, None)
            if char == line.text.strip():
                if not self.key_phrases_built[i]:
                    self.key_phrase_beginning_line[i] = line
                self.key_phrases_built[i] += char
                if self.key_phrases_built[i] == ''.join(self.key_phrases[i].split()):
                    return self.key_phrase_beginning_line[i]
            else:
                self.key_phrases_built[i] = ''
                self.key_phrases_iter[i] = self.iter_key_phrase(self.key_phrases[i])
                self.key_phrase_beginning_line[i] = None
        return None

def serialize_lines_with_pointers(lines: list[PatentTextLine]):
    """
    Serialize column-interleaved text lines and return a mapping of pointers to indices in the serialized string for the original patent text lines
    """
    serialized_lines = ''
    pointers = []
    for line in lines:
        pointers.append(len(serialized_lines))
        serialized_lines += line.text
    return serialized_lines, pointers

# def find_end_of_claims(column_0: list[PatentTextLine], column_1: list[PatentTextLine]) -> EndOfClaims:
#     """ 
#     Returns page number and line number where claims end
#     """
#     end_claims_classification_model = EndClaimsClassificationModel()

#     # Check column 0
#     for text_line in column_0:
#         if text_line.page_num == 16:
#             print(text_line.text)
#         found_line = end_claims_classification_model.classify(text_line)
#         if found_line:
#             return EndOfClaims(page_index=found_line.page_num, col_index=0, line_index=found_line.line_index)
        
#     # Check column 1
#     for text_line in column_1:
#         if text_line.page_num == 16:
#             print(text_line.text)
#         found_line = end_claims_classification_model.classify(text_line)
#         if found_line:
#             return EndOfClaims(page_index=found_line.page_num, col_index=1, line_index=found_line.line_index)
#     return None

def find_end_of_claims(column_0: list[PatentTextLine], column_1: list[PatentTextLine], beginning_of_claims: BeginningOfClaims) -> EndOfClaims | None:
    column_interleaved = interleave_patent_text_lines_columns(column_0, column_1)
    # Extract chunk after beginning of claims
    interleaved_claims_start_line = None
    for i in range(len(column_interleaved)):
        if (column_interleaved[i].page_num == beginning_of_claims.page_index 
        and column_interleaved[i].column_index == beginning_of_claims.col_index and column_interleaved[i].line_index == beginning_of_claims.line_index):
            interleaved_claims_start_line = i
            break

    claim_lines = column_interleaved[interleaved_claims_start_line:]
    serialized_lines, pointers = serialize_lines_with_pointers(claim_lines)

    pointers_index = 0
    for i, char in enumerate(serialized_lines):
        if (
            char == '.' 
            and not re.match(claim_number_regex, serialized_lines[pointers[pointers_index]:i+1])  # This period isn't part of the claim's numbering
            and not re.match(claim_number_regex, serialized_lines[pointers[pointers_index+1]:])  # The nThe system of claim 1 , wherein the user input is a firstext line doesn't have claim numbering
            and all(
                c.isspace() for c in claim_lines[pointers_index].text[i - pointers[pointers_index]+1:]
            )  # Check period is at end of the line (sometimes there are spurious periods mid line. If there's a spurious period at the end of a line, we're out of luck)
        ):
            return EndOfClaims(
                page_index=claim_lines[pointers_index].page_num,
                col_index=claim_lines[pointers_index].column_index,
                line_index=claim_lines[pointers_index].line_index
            )

        if pointers_index < len(pointers) - 1 and i == pointers[pointers_index + 1]:
            pointers_index += 1  

    return None

@dataclass
class PatentTextLines:
    column_0: list[PatentTextLine]
    column_1: list[PatentTextLine]
    noncolumnar: list[PatentTextLine]

class ParsedPatent(pydantic.BaseModel):
    unique_id: PatentUniqueID
    text_lines: PatentTextLines
    beginning_of_specification: int
    # Currently optional to accomodate cases where claims don't have clear prefatory language (see example in tests). Should be made not optional
    beginning_of_claims: Optional[BeginningOfClaims]
    end_of_claims: Optional[EndOfClaims]

class GoogleParsedPatent(pydantic.BaseModel):
    unique_id: PatentUniqueID
    specification: str
    claims: str

is_column_num_pattern = r'^\d{1,4}\s*(?:\n|\\n)?\s*$'
is_column_num_regex = re.compile(is_column_num_pattern)

paragraph_num_pattern = r'\s*\[\s*(?P<paragraph_num>\d{4})\s*\]'
paragraph_num_regex = re.compile(paragraph_num_pattern)

def parse_patent_unique_id_from_text_lines(text_lines: TextLines) -> PatentUniqueID | None:
    column_0 = text_lines.column_0
    column_1 = text_lines.column_1
    noncolumnar = text_lines.noncolumnar

    # Search for unique id in the first page of the patent. Start with column 1, because that's usually where it is
    unique_id = None
    for line in column_1:
        # gd pdfs
        text = re.sub(r'\s+', '', line.text)
        text = text.replace('l', '1')
        match = patent_number_regex.search(text)
        if match:
            patent_number = match.groupdict()['patent_number']
            year = match.groupdict()['year']
            if year:
                patent_number = year + patent_number
            kind_code = match.groupdict()['kind_code']
            if kind_code:
                kind_code = kind_code.replace('l', '1')
            unique_id = PatentUniqueID(
                patent_number=patent_number,
                country_code=match.groupdict()['country_code'],
                kind_code=kind_code
            )
            break

    if not unique_id:
        for line in column_0:
            # gd pdfs
            text = re.sub(r'\s+', '', line.text)
            text = text.replace('l', '1')
            match = patent_number_regex.search(text)
            if match:
                patent_number = match.groupdict()['patent_number']
                year = match.groupdict()['year']
                if year:
                    patent_number = year + patent_number
                kind_code = match.groupdict()['kind_code']
                if kind_code:
                    kind_code = kind_code.replace('l', '1')
                unique_id = PatentUniqueID(
                    patent_number=patent_number,
                    country_code=match.groupdict()['country_code'],
                    kind_code=kind_code
                )
                break
    
    if not unique_id:
        for line in noncolumnar:
            # gd pdfs
            text = re.sub(r'\s+', '', line.text)
            text = text.replace('l', '1')
            match = patent_number_regex.search(text)
            if match:
                patent_number = match.groupdict()['patent_number']
                year = match.groupdict()['year']
                if year:
                    patent_number = year + patent_number
                kind_code = match.groupdict()['kind_code']
                if kind_code:
                    kind_code = kind_code.replace('l', '1')
                unique_id = PatentUniqueID(
                    patent_number=patent_number,
                    country_code=match.groupdict()['country_code'],
                    kind_code=kind_code
                )
                break

    return unique_id

def parse_patent_unique_id_from_pdf_path(path: pathlib.Path | str) -> PatentUniqueID | None:
    text_lines = parse_text_lines_from_pdf_path_simple(path, pages=[0])
    return parse_patent_unique_id_from_text_lines(text_lines)

def parse_patent_unique_id_from_pdf_file_obj(f: BinaryIO) -> PatentUniqueID | None:
    text_lines = parse_text_lines_from_pdf_file_obj_simple(f, pages=[0])
    return parse_patent_unique_id_from_text_lines(text_lines)

def parse_patent_from_text_lines(text_lines: TextLines) -> ParsedPatent:
    """ 
    Assumes all arguments sorted first by page number ascending
    """
    column_0 = text_lines.column_0
    column_1 = text_lines.column_1
    noncolumnar = text_lines.noncolumnar

    column_0_patent_text_lines = []
    column_1_patent_text_lines = []
    noncolumnar_patent_text_lines = []

    def is_column_num(line: TextLine, line_index: int):
        return (
            line_index == 0 and
            is_column_num_regex.match(line.text)
        )

    def get_line_num(line: TextLine, line_index: int) -> int:
        """
        If line contains a line number, returns the line number and trims the line number from the line. Otherwise returns None
        """
        # TODO: unify this with the parsing done by the PDFPageDetailedAggregator
        # TODO: currently misses a lot of line numbers because many line numbers are parsed as their own 1-2 character line. Associating the line number with the appropriate line will have to happen upstream, perhaps by appending the line number to the line closest in alignment on the y axis
        if line_index == 0:
            return None
        if len(line.text) < 3:
            # Currently, columnar lines include random text from figures and tables. These lines will be given line numbers, making this function noisy, but we add this simple check just to make sure this function can run without error. We know lines we care about will be at least 3 characters long.
            return None
        candidate_line_num = 0
        # Line number is at beginning of line
        first_char = line.text[0]
        if first_char.isdigit():
            candidate_line_num = first_char
            second_char = line.text[1]
            if second_char.isdigit():
                if line.text[2] == '.':
                    # This is a probably a claim number
                    return None
                candidate_line_num = candidate_line_num + second_char
            elif second_char == '.':
                # This is a probably a claim number
                return None

            # If candidate line number is a multiple of 5, then it is a line number
            if int(candidate_line_num) % 5 == 0:
                # Strip line text of candidate line num
                # TODO: fix because this erroneously strips claim numbers sometimes
                # line.text = line.text[len(candidate_line_num):]
                return int(candidate_line_num)
            
        # Line number is at end of line
        last_char = line.text.rstrip('\n')[-1]
        if last_char.isdigit():
            candidate_line_num = last_char
            second_to_last_char = line.text.rstrip('\n')[-2]
            if second_to_last_char.isdigit():
                candidate_line_num = second_to_last_char + candidate_line_num
            # If candidate line number is a multiple of 5, then it is a line number
            if int(candidate_line_num) % 5 == 0:
                # Strip line text of candidate line num
                # TODO: fix because this erroneously strips claim numbers sometimes
                # line.text = line.text[:-(len(candidate_line_num))]
                return int(candidate_line_num)
        return None

    def get_paragraph_num(line: TextLine):
        match = paragraph_num_regex.match(line.text)
        if match:
            return int(match.groupdict()['paragraph_num'])
        return None

    # Search for unique id in the first page of the patent. Start with column 1, because that's usually where it is
    unique_id = None
    for line in column_1:
        # gd pdfs
        text = re.sub(r'\s+', '', line.text)
        text = text.replace('l', '1')
        match = patent_number_regex.search(text)
        if match:
            patent_number = match.groupdict()['patent_number']
            year = match.groupdict()['year']
            if year:
                patent_number = year + patent_number
            kind_code = match.groupdict()['kind_code']
            if kind_code:
                kind_code = kind_code.replace('l', '1')
            unique_id = PatentUniqueID(
                patent_number=patent_number,
                country_code=match.groupdict()['country_code'],
                kind_code=kind_code
            )
            break

    if not unique_id:
        for line in column_0:
            # gd pdfs
            text = re.sub(r'\s+', '', line.text)
            text = text.replace('l', '1')
            match = patent_number_regex.search(text)
            if match:
                patent_number = match.groupdict()['patent_number']
                year = match.groupdict()['year']
                if year:
                    patent_number = year + patent_number
                kind_code = match.groupdict()['kind_code']
                if kind_code:
                    kind_code = kind_code.replace('l', '1')
                unique_id = PatentUniqueID(
                    patent_number=patent_number,
                    country_code=match.groupdict()['country_code'],
                    kind_code=kind_code
                )
                break
    
    if not unique_id:
        for line in noncolumnar:
            # gd pdfs
            text = re.sub(r'\s+', '', line.text)
            text = text.replace('l', '1')
            match = patent_number_regex.search(text)
            if match:
                patent_number = match.groupdict()['patent_number']
                year = match.groupdict()['year']
                if year:
                    patent_number = year + patent_number
                kind_code = match.groupdict()['kind_code']
                if kind_code:
                    kind_code = kind_code.replace('l', '1')
                unique_id = PatentUniqueID(
                    patent_number=patent_number,
                    country_code=match.groupdict()['country_code'],
                    kind_code=kind_code
                )
                break

    line_index = 0
    current_page_num = 0
    column_num = None
    paragraph_num = None
    for line in column_0:
        page_num = line.page_number
        if page_num != current_page_num:
            line_index = 0
            current_page_num = page_num
        if is_column_num(line, line_index):
            column_num = int(line.text.rstrip('\n'))
            if column_num % 2 == 0:
                logger.error(
                    f'Column number {column_num} unexpectedly even. Patent: {unique_id} -- Line: {line}'
                )
                column_num = None
            
            # NOTE: don't include column number lines in the parsed text
            continue

        if line.text.strip().isdigit() and int(line.text.strip()) % 5 == 0:
            # TODO: for now we just throw away standalone line numbers. See comment in get_line_num
            continue

        # TODO: currently only adds line number to line that has the numbering. Should add line number to each line based on the line's offset from the last line with an explicit numbering
        line_num = get_line_num(line, line_index)
        new_paragraph_num = get_paragraph_num(line)
        if new_paragraph_num is not None:
            paragraph_num = new_paragraph_num
        column_0_patent_text_lines.append(
            PatentTextLine(
                page_num=line.page_number, 
                text=line.text, 
                line_index=line_index, 
                column_index=0,
                column_num=column_num,
                line_num=line_num,
                paragraph_num=paragraph_num
            )
        )

        line_index += 1

    line_index = 0
    current_page_num = 0
    column_num = None
    paragraph_num = None
    for line in column_1:
        page_num = line.page_number
        if page_num != current_page_num:
            line_index = 0
            current_page_num = page_num
        if is_column_num(line, line_index):
            column_num = int(line.text.rstrip('\n'))
            if column_num % 2 == 1:
                logger.error(
                    f'Column number {column_num} unexpectedly odd. Patent: {unique_id} -- Line: {line}'
                )
                column_num = None
            
            # NOTE: don't include column number lines in the parsed text
            continue

        if line.text.strip().isdigit() and int(line.text.strip()) % 5 == 0:
            # TODO: for now we just throw away standalone line numbers. See comment in get_line_num
            continue

        line_num = get_line_num(line, line_index)
        paragraph_num = get_paragraph_num(line)
        if new_paragraph_num is not None:
            paragraph_num = new_paragraph_num
        column_1_patent_text_lines.append(
            PatentTextLine(
                page_num=line.page_number, 
                text=line.text, 
                line_index=line_index, 
                column_index=1,
                column_num=column_num,
                line_num=line_num,
                paragraph_num=paragraph_num
            )
        )

        line_index += 1

    line_index = 0
    current_page_num = 0
    for line in noncolumnar:
        page_num = line.page_number
        if page_num != current_page_num:
            line_index = 0
            current_page_num = page_num
        noncolumnar_patent_text_lines.append(
            PatentTextLine(
                page_num=line.page_number, 
                text=line.text, 
                line_index=line_index, 
                column_index=None,
                column_num=None,
                line_num=None,
                paragraph_num=None
            )
        )

        line_index += 1
    
    beginning_of_claims = find_beginning_of_claims(column_0_patent_text_lines, column_1_patent_text_lines)
    end_of_claims = find_end_of_claims(column_0_patent_text_lines, column_1_patent_text_lines, beginning_of_claims)

    return ParsedPatent(
        unique_id=unique_id,
        text_lines=PatentTextLines(
            column_0=column_0_patent_text_lines,
            column_1=column_1_patent_text_lines,
            noncolumnar=noncolumnar_patent_text_lines,
        ),
        beginning_of_specification=find_beginning_of_specification(column_0_patent_text_lines, column_1_patent_text_lines),
        beginning_of_claims=beginning_of_claims,
        end_of_claims=end_of_claims
    )

def parse_patent_from_pdf_path(path: pathlib.Path | str) -> ParsedPatent:
    text_lines = parse_text_lines_from_pdf_path(path)
    return parse_patent_from_text_lines(text_lines)

def parse_patent_from_pdf_file_obj(f: BinaryIO) -> ParsedPatent | None:
    text_lines = parse_text_lines_from_pdf_file_obj(f)
    return parse_patent_from_text_lines(text_lines)

def parse_google_patent_from_pdf_path(path: pathlib.Path | str) -> GoogleParsedPatent:
    unique_id = parse_patent_unique_id_from_pdf_path(path)
    if unique_id is None:
        raise ParseError(f'Could not parse unique id from patent')
    google_unique_id = f'{unique_id.country_code}{unique_id.patent_number}{unique_id.kind_code or ""}'
    google_patent_result = parse_google_patent(google_unique_id)
    if google_patent_result is not None:
        specification, claims = google_patent_result
        return GoogleParsedPatent(
            unique_id=unique_id,
            specification=specification,
            claims=claims
        )
    else:
        raise ParseError(f'Could not parse google patent for unique id {google_unique_id}')

def parse_google_patent_from_pdf_file_obj(f: BinaryIO) -> GoogleParsedPatent:
    unique_id = parse_patent_unique_id_from_pdf_file_obj(f)
    if unique_id is None:
        raise ParseError(f'Could not parse unique id from patent')
    google_unique_id = f'{unique_id.country_code}{unique_id.patent_number}{unique_id.kind_code or ""}'
    google_patent_result = parse_google_patent(google_unique_id)
    if google_patent_result is not None:
        specification, claims = google_patent_result
        return GoogleParsedPatent(
            unique_id=unique_id,
            specification=specification,
            claims=claims
        )
    else:
        raise ParseError(f'Could not parse google patent for unique id {google_unique_id}')

def parse_google_patent_from_unique_id(unique_id: str) -> GoogleParsedPatent:
    unique_id = unique_id.replace('-', '')
    match = patent_number_regex.search(unique_id)
    parsed_unique_id = None
    if match:
        patent_number = match.groupdict()['patent_number']
        year = match.groupdict()['year']
        if year:
            patent_number = year + patent_number
        kind_code = match.groupdict()['kind_code']
        parsed_unique_id = PatentUniqueID(
            patent_number=patent_number,
            country_code=match.groupdict()['country_code'],
            kind_code=kind_code
        )
    else:
        raise ParseError(f'Could not parse unique id {unique_id}')
    google_patent_result = parse_google_patent(unique_id)
    if google_patent_result is not None:
        specification, claims = google_patent_result
        return GoogleParsedPatent(
            unique_id=parsed_unique_id,
            specification=specification,
            claims=claims,
        )
    else:
        raise ParseError(f'Could not parse google patent for unique id {unique_id}')

def interleave_patent_text_lines_columns(column_0: list[PatentTextLine], column_1: list[PatentTextLine]) -> list[PatentTextLine]:
    """
    Assumes column_0 and column_1 are sorted by page number, then by line index
    """
    interleaved = []
    i = 0
    j = 0
    while i < len(column_0) and j < len(column_1):
        if column_0[i].page_num <= column_1[j].page_num:
            interleaved.append(column_0[i])
            i += 1
        else:
            interleaved.append(column_1[j])
            j += 1
    return interleaved

def group_parsed_patent_by_page(parsed_patent: ParsedPatent) -> dict[int, PatentTextLines]:
    pages_to_patent_text_lines = {}
    for text_line in parsed_patent.text_lines.column_0:
        if text_line.page_num not in pages_to_patent_text_lines:
            pages_to_patent_text_lines[text_line.page_num] = PatentTextLines(
                column_0=[],
                column_1=[],
                noncolumnar=[]
            )
        pages_to_patent_text_lines[text_line.page_num].column_0.append(text_line)

    for text_line in parsed_patent.text_lines.column_1:
        if text_line.page_num not in pages_to_patent_text_lines:
            pages_to_patent_text_lines[text_line.page_num] = PatentTextLines(
                column_0=[],
                column_1=[],
                noncolumnar=[]
            )
        pages_to_patent_text_lines[text_line.page_num].column_1.append(text_line)

    for text_line in parsed_patent.text_lines.noncolumnar:
        if text_line.page_num not in pages_to_patent_text_lines:
            pages_to_patent_text_lines[text_line.page_num] = PatentTextLines(
                column_0=[],
                column_1=[],
                noncolumnar=[]
            )
        pages_to_patent_text_lines[text_line.page_num].noncolumnar.append(text_line)

    return pages_to_patent_text_lines

@dataclass
class PatentWord:
    text:str
    line: PatentTextLine

def flatten_parsed_patent_to_words(parsed_patent: ParsedPatent) -> list[PatentWord]:
    """
    Returns a list of words from the columnar lines of the patent, ordered by page then by column then by line index.
    """
    lines_by_page = group_parsed_patent_by_page(parsed_patent)
    words: list[PatentWord] = []
    for _, lines in lines_by_page.items():
        for line in lines.column_0:
            words.extend(
                PatentWord(
                    text=word,
                    line=line,
                )
                for word in line.text.split()
            )
        for line in lines.column_1:
            words.extend(
                PatentWord(
                    text=word,
                    line=line,
                )
                for word in line.text.split()
            )
    return words

def serialize_claims_from_parsed_patent(parsed_patent: ParsedPatent) -> list[str]:
    """
    Extract text from beginning of claims to end of claims.
    """
    by_page = group_parsed_patent_by_page(parsed_patent)
    serialized_claims = []
    claim_beg_page = parsed_patent.beginning_of_claims.page_index
    claim_beg_col = parsed_patent.beginning_of_claims.col_index
    # Add one because we don't want to include the prefatory text
    claim_beg_line = parsed_patent.beginning_of_claims.line_index + 1
    claim_end_page = parsed_patent.end_of_claims.page_index
    claim_end_col = parsed_patent.end_of_claims.col_index
    claim_end_line = parsed_patent.end_of_claims.line_index

    # Get text from beginning to end
    if claim_beg_page == claim_end_page:
        page = by_page[claim_beg_page]
        # Get text from beginning of claims to end of claims
        if claim_beg_col == 0 and claim_end_col == 0:
            # Get text from beginning in column 0 to end in column 0
            for text_line in page.column_0[claim_beg_line:claim_end_line]:
                serialized_claims.append(text_line.text)
        elif claim_beg_col == 0:
            # Get text from beginning in column 0 to end in column 1
            for text_line in page.column_0[claim_beg_line:]:
                serialized_claims.append(text_line.text)
            for text_line in page.column_1[:claim_end_line]:
                serialized_claims.append(text_line.text)
        else:
            # Get text from beginning in column 1 to end in column 1
            for text_line in page.column_1[claim_beg_line:claim_end_line]:
                serialized_claims.append(text_line.text)
    else:
        for page_num in range(claim_beg_page, claim_end_page + 1):
            page = by_page[page_num]
            if page_num == claim_beg_page:
                # Get text from beginning of claims to end of page
                if claim_beg_col == 0:
                    # Get text from beginning of claims to end of column 0
                    for text_line in page.column_0[claim_beg_line:]:
                        serialized_claims.append(text_line.text)
                    for text_line in page.column_1:
                        serialized_claims.append(text_line.text)
                else:
                    # Get text from beginning of claims to end of column 1
                    for text_line in page.column_1[claim_beg_line:]:
                        serialized_claims.append(text_line.text)
            elif page_num == claim_end_page:
                # Get text from beginning of page to end of claims
                if claim_end_col == 0:
                    # Get text from beginning of column 0 to end of claims
                    for text_line in page.column_0[:claim_end_line]:
                        serialized_claims.append(text_line.text)
                else:
                    # Get text from beginning of column 1 to end of claims
                    for text_line in page.column_0:
                        serialized_claims.append(text_line.text)
                    for text_line in page.column_1[:claim_end_line]:
                        serialized_claims.append(text_line.text)
            else:
                # Get text from beginning of page to end of page
                for text_line in page.column_0:
                    serialized_claims.append(text_line.text)
                for text_line in page.column_1:
                    serialized_claims.append(text_line.text)

    return serialized_claims

def serialize_specification_from_parsed_patent(parsed_patent: ParsedPatent) -> list[str]:
    """
    Extract the specifiction from the parsed patent. We assume the specification ends at the beginning of the claims. We assume the specification starts at the beginning of the first page of the specification. We assume the specification starts at the beginning of the first column of the first page of the specification.
    """
    by_page = group_parsed_patent_by_page(parsed_patent)
    serialized_spec = []
    beg_spec_page = parsed_patent.beginning_of_specification
    beg_spec_line = 0
    claim_beg_page = parsed_patent.beginning_of_claims.page_index
    claim_beg_col = parsed_patent.beginning_of_claims.col_index
    claim_beg_line = parsed_patent.beginning_of_claims.line_index
    # We'll allow claim_beg_page to be None if the claims are not found. In this case, we'll just serialize the entire specification, including the claims.
    if claim_beg_page is None:
        claim_beg_page = len(by_page) - 1

    # Get text from beginning to end
    for page_num in range(beg_spec_page, claim_beg_page + 1):
        page = by_page[page_num]
        if page_num == beg_spec_page:
            # Get text from beginning of specification to end of page. We assume spec begins at the beginning of col 0 of the first page of the spec.
            for text_line in page.column_0[beg_spec_line:]:
                serialized_spec.append(text_line.text)
            for text_line in page.column_1:
                serialized_spec.append(text_line.text)
        elif page_num == claim_beg_page:
            # Get text from beginning of page to beginning of claims
            if claim_beg_col == 0:
                # Get text from beginning of column 0 to beginning of claims
                for text_line in page.column_0[:claim_beg_line]:
                    serialized_spec.append(text_line.text)
            else:
                # Get text from beginning of column 1 to beginning of claims
                for text_line in page.column_0:
                    serialized_spec.append(text_line.text)
                for text_line in page.column_1[:claim_beg_line]:
                    serialized_spec.append(text_line.text)
        else:
            # Get text from beginning of page to end of page
            for text_line in page.column_0:
                serialized_spec.append(text_line.text)
            for text_line in page.column_1:
                serialized_spec.append(text_line.text)

    return serialized_spec

def parse_claims_from_parsed_patent(parsed_patent: ParsedPatent) -> Claims:
    serialized_claims = serialize_claims_from_parsed_patent(parsed_patent)
    claim_text = ''.join(serialized_claims)
    tokens = lex_claims(claim_text)
    parser = ClaimsParser(tokens)
    return parser.parse()

def parse_claims_from_google_parsed_patent(parsed_patent: GoogleParsedPatent) -> list[GoogleClaim]:
    # Trim beginning of claims
    claim_begin = parsed_patent.claims.find('1.')

    trimmed_claims = parsed_patent.claims[claim_begin:]
    claim_lines = [l.strip() for l in trimmed_claims.split('\n') if l.strip()]

    claim_number = 1
    claims = []
    claim = GoogleClaim(
        claim_number=claim_number,
        claim_elements=[]
    )
    for line in claim_lines:
        if line.startswith(str(claim_number + 1)):
            claim_number += 1
            claims.append(claim)
            claim = GoogleClaim(
                claim_number=claim_number,
                claim_elements=[]
            )
        claim.claim_elements.append(line)

    return claims

# def chunk_spec_from_parsed_patent(parsed_patent, chunk_word_count=250):
#     by_page = group_parsed_patent_by_page(parsed_patent)
#     chunks = []
#     beg_spec_page = parsed_patent.beginning_of_specification
#     beg_spec_line = 0
#     claim_beg_page = parsed_patent.beginning_of_claims.page_index
#     claim_beg_col = parsed_patent.beginning_of_claims.col_index
#     claim_beg_line = parsed_patent.beginning_of_claims.line_index
#     # We'll allow claim_beg_page to be None if the claims are not found. In this case, we'll just serialize the entire specification, including the claims.
#     if claim_beg_page is None:
#         claim_beg_page = len(by_page) - 1

#     def count_words_in_chunk(chunk):
#         return sum([len(text_line.text.split()) for text_line in chunk])

#     chunk = []
#     # Get text from beginning to end
#     for page_num in range(beg_spec_page, claim_beg_page + 1):
#         page = by_page[page_num]
#         if page_num == beg_spec_page:
#             # Get text from beginning of specification to end of page. We assume spec begins at the beginning of col 0 of the first page of the spec.
#             for text_line in page.column_0[beg_spec_line:]:
#                 chunk.append(text_line)
#                 if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                     chunks.append(chunk)
#                     chunk = []
#             for text_line in page.column_1:
#                 chunk.append(text_line)
#                 if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                     chunks.append(chunk)
#                     chunk = []
#         elif page_num == claim_beg_page:
#             # Get text from beginning of page to beginning of claims
#             if claim_beg_col == 0:
#                 # Get text from beginning of column 0 to beginning of claims
#                 for text_line in page.column_0[:claim_beg_line]:
#                     chunk.append(text_line)
#                     if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                         chunks.append(chunk)
#                         chunk = []
#             else:
#                 # Get text from beginning of column 1 to beginning of claims
#                 for text_line in page.column_0:
#                     chunk.append(text_line)
#                     if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                         chunks.append(chunk)
#                         chunk = []
#                 for text_line in page.column_1[:claim_beg_line]:
#                     chunk.append(text_line)
#                     if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                         chunks.append(chunk)
#                         chunk = []
#         else:
#             # Get text from beginning of page to end of page
#             for text_line in page.column_0:
#                 chunk.append(text_line)
#                 if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                         chunks.append(chunk)
#                         chunk = []
#             for text_line in page.column_1:
#                 chunk.append(text_line)
#                 if count_words_in_chunk(chunk) >= chunk_word_count and text_line.text.endswith('.'):
#                         chunks.append(chunk)
#                         chunk = []

#     return chunks

def chunk_spec_from_parsed_patent(parsed_patent, chunk_word_count=250):
    """
    Splits on a period that occurs anywhere within the text line rather than just at the end, leading to more accurate word count chunks.
    """
    by_page = group_parsed_patent_by_page(parsed_patent)
    chunks = []
    beg_spec_page = parsed_patent.beginning_of_specification
    beg_spec_line = 0
    claim_beg_page = parsed_patent.beginning_of_claims.page_index
    claim_beg_col = parsed_patent.beginning_of_claims.col_index
    claim_beg_line = parsed_patent.beginning_of_claims.line_index
    # We'll allow claim_beg_page to be None if the claims are not found. In this case, we'll just serialize the entire specification, including the claims.
    if claim_beg_page is None:
        claim_beg_page = len(by_page) - 1

    def count_words_in_chunk(chunk):
        return sum([len(text_line.text.split()) for text_line in chunk])

    chunk = []
    # Get text from beginning to end
    for page_num in range(beg_spec_page, claim_beg_page + 1):
        page = by_page[page_num]
        if page_num == beg_spec_page:
            # Get text from beginning of specification to end of page. We assume spec begins at the beginning of col 0 of the first page of the spec.
            for text_line in page.column_0[beg_spec_line:]:
                if '.' in text_line.text:
                    # split on period into two text lines
                    split_text = text_line.text.split('.')
                    # copy text line and replace text with first half of split
                    text_line_before_period = copy.copy(text_line)
                    text_line_before_period.text = split_text[0] + '.'
                    text_line_after_period = copy.copy(text_line)
                    text_line_after_period.text = split_text[1].lstrip()

                    if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                        chunk.append(text_line_before_period)
                        chunks.append(chunk)
                        chunk = []
                        if text_line_after_period.text:
                            chunk.append(text_line_after_period)
                else:
                    chunk.append(text_line)
            for text_line in page.column_1:
                if '.' in text_line.text:
                    # split on period into two text lines
                    split_text = text_line.text.split('.')
                    # copy text line and replace text with first half of split
                    text_line_before_period = copy.copy(text_line)
                    text_line_before_period.text = split_text[0] + '.'
                    text_line_after_period = copy.copy(text_line)
                    text_line_after_period.text = split_text[1].lstrip()

                    if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                        chunk.append(text_line_before_period)
                        chunks.append(chunk)
                        chunk = []
                        if text_line_after_period.text:
                            chunk.append(text_line_after_period)
                else:
                    chunk.append(text_line)
        elif page_num == claim_beg_page:
            # Get text from beginning of page to beginning of claims
            if claim_beg_col == 0:
                # Get text from beginning of column 0 to beginning of claims
                for text_line in page.column_0[:claim_beg_line]:
                    if '.' in text_line.text:
                        # split on period into two text lines
                        split_text = text_line.text.split('.')
                        # copy text line and replace text with first half of split
                        text_line_before_period = copy.copy(text_line)
                        text_line_before_period.text = split_text[0] + '.'
                        text_line_after_period = copy.copy(text_line)
                        text_line_after_period.text = split_text[1].lstrip()

                        if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                            chunk.append(text_line_before_period)
                            chunks.append(chunk)
                            chunk = []
                            if text_line_after_period.text:
                                chunk.append(text_line_after_period)
                    else:
                        chunk.append(text_line)
            else:
                # Get text from beginning of column 1 to beginning of claims
                for text_line in page.column_0:
                    if '.' in text_line.text:
                        # split on period into two text lines
                        split_text = text_line.text.split('.')
                        # copy text line and replace text with first half of split
                        text_line_before_period = copy.copy(text_line)
                        text_line_before_period.text = split_text[0] + '.'
                        text_line_after_period = copy.copy(text_line)
                        text_line_after_period.text = split_text[1].lstrip()

                        if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                            chunk.append(text_line_before_period)
                            chunks.append(chunk)
                            chunk = []
                            if text_line_after_period.text:
                                chunk.append(text_line_after_period)
                    else:
                        chunk.append(text_line)
                for text_line in page.column_1[:claim_beg_line]:
                    if '.' in text_line.text:
                        # split on period into two text lines
                        split_text = text_line.text.split('.')
                        # copy text line and replace text with first half of split
                        text_line_before_period = copy.copy(text_line)
                        text_line_before_period.text = split_text[0] + '.'
                        text_line_after_period = copy.copy(text_line)
                        text_line_after_period.text = split_text[1].lstrip()

                        if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                            chunk.append(text_line_before_period)
                            chunks.append(chunk)
                            chunk = []
                            if text_line_after_period.text:
                                chunk.append(text_line_after_period)
                    else:
                        chunk.append(text_line)
        else:
            # Get text from beginning of page to end of page
            for text_line in page.column_0:
                if '.' in text_line.text:
                    # split on period into two text lines
                    split_text = text_line.text.split('.')
                    # copy text line and replace text with first half of split
                    text_line_before_period = copy.copy(text_line)
                    text_line_before_period.text = split_text[0] + '.'
                    text_line_after_period = copy.copy(text_line)
                    text_line_after_period.text = split_text[1].lstrip()

                    if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                        chunk.append(text_line_before_period)
                        chunks.append(chunk)
                        chunk = []
                        if text_line_after_period.text:
                            chunk.append(text_line_after_period)
                else:
                    chunk.append(text_line)
            for text_line in page.column_1:
                if '.' in text_line.text:
                    # split on period into two text lines
                    split_text = text_line.text.split('.')
                    # copy text line and replace text with first half of split
                    text_line_before_period = copy.copy(text_line)
                    text_line_before_period.text = split_text[0] + '.'
                    text_line_after_period = copy.copy(text_line)
                    text_line_after_period.text = split_text[1].lstrip()

                    if count_words_in_chunk(chunk) + len(text_line.text.split()) >= chunk_word_count:
                        chunk.append(text_line_before_period)
                        chunks.append(chunk)
                        chunk = []
                        if text_line_after_period.text:
                            chunk.append(text_line_after_period)
                else:
                    chunk.append(text_line)

    return chunks

def serialize_chunk(chunk):
    return ''.join([text_line.text for text_line in chunk])


from dataclasses import dataclass
from typing import Optional

@dataclass
class PatentTextLine:
    page_num: int  # 0-based index returned from PDFPageDetailedAggregator. TODO: rename to index
    text: str
    line_index: int  # line index counting TextLine instances from top to bottom of page
    column_index: Optional[int]  # 0 or 1 or None if noncolumnar
    line_num: Optional[int] = None  # human labeled line number, multiple of 5
    column_num: Optional[int] = None  # human labeled column number
    paragraph_num: Optional[int] = None  # alternative human labeling method often in form [XXXX] at beginning of paragraph

@dataclass
class GeneratedPassage:
    prior_art_source_id: int
    text: str
    claim_element_id: int
    model_id: str
    start_line: Optional[PatentTextLine] = None
    end_line: Optional[PatentTextLine] = None
    ranking: int = 0


import re
import time
import asyncio
import uuid
import random
from logging import getLogger
from dataclasses import dataclass
from typing import Callable, Optional, AsyncIterator, Coroutine
from functools import partial

import openai
import tiktoken
from dotenv import load_dotenv

from . import parser
from . import ranking
from .utils import num_tokens_from_messages
from .data_structures import GeneratedPassage
from .requests_utils import DistributedFixedWindowRateLimiter

logger = getLogger(__name__)

load_dotenv()

# _openai_sync_client = None
# _openai_async_client = None

# def openai_sync_client():
#     global _openai_sync_client
#     if _openai_sync_client is None:
#         _openai_sync_client = openai.OpenAI()
#     return _openai_sync_client

# def openai_async_client():
#     global _openai_async_client
#     if _openai_async_client is None:
#         _openai_async_client = openai.AsyncOpenAI()
#     return _openai_async_client

class GenerationError(Exception):
    pass

RATE_LIMIT_INTERVAL = 60  # seconds

token_rate_limit_by_model_name = {
    'gpt-4': 300_000,
    'gpt-3.5-turbo': 1_000_000,
    'gpt-3.5-turbo-16k': 1_000_000,
    'gpt-3.5-turbo-instruct': 250_000,
}

model_name_to_token_limit = {
    'gpt-4': 8192,
    'gpt-3.5-turbo-16k': 16384,
    'gpt-3.5-turbo': 4096,
    'gpt-3.5-turbo-16k-0613': 16384,
}

def prompt_claim_element_only(claim_element, prior_art_passage, n_passages_per_split=3):
    messages = [
        {"role": "system", "content": f"You work for a law firm. Your task is to interpret a single claim element that I select from the claims of a patent. Then I'm going to give you a passage from the prior art and you need to pick {n_passages_per_split} distinct sub-passages from the passage that you believe are the most similar to the claim element. Please output only those {n_passages_per_split} passages, copied verbatim from the prior art. Do not add any other text besides the verbatim passages."},
        {"role": "user", "content": f"Selected claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passage: {prior_art_passage}"},
    ]
    return messages

def prompt_summary_full_set_of_claims(summary, claims, claim_element, prior_art_passage, n_passages_per_split=3):
    messages = [
        {"role": "system", "content": f"You work for a law firm. Your task is to read the summary of the written description of a patent and the full set of claims in the patent and interpret a single claim element that I select from the claims of the patent in light of the summary and the full set of claims. Then I'm going to give you a passage from the prior art and you need to pick {n_passages_per_split} distinct sub-passages from the passage that you believe are the most similar to the claim element. Please output only those {n_passages_per_split} passages, copied verbatim from the prior art. Do not add any other text besides the verbatim passages."},
        {"role": "user", "content": f"Summary of the written description of the invention: {summary}"},
        {"role": "user", "content": f"Full set of claims: {claims}"},
        {"role": "user", "content": f"Selected claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passage: {prior_art_passage}"},
    ]
    return messages

def post_process_selected_passage_gpt4(passage: str) -> list[str]:
    split_strategies = [
        lambda passage: re.split(r'\d+\.', passage),
        lambda passage: passage.split('\n\n'),
        lambda passage: passage.split('\"\n'),
    ]
    for split_strategy in split_strategies:
        split_passages = split_strategy(passage)
        if len(split_passages) > 1:
            break
    # strip newlines, quotes, and whitespace from beginning and end
    split_passages = [p.strip().strip('"') for p in split_passages]
    # filter whitespace only and empty string entries
    split_passages = [p for p in split_passages if p != '']
    # filter out passages that are just a newline
    split_passages = [p for p in split_passages if p != '\n']
        
    return split_passages

@dataclass
class GeneratorConfig:
    model_name: str = 'gpt-4'
    target_n_passages: int = 10
    min_n_splits: int = 1
    generate_passages_prompt_func: Callable[..., list[dict]] = prompt_summary_full_set_of_claims
    post_process_generated_passages_func: Callable[[str], list[str]] = post_process_selected_passage_gpt4

def load_generator_config_from_env():
    return GeneratorConfig()

def num_tokens_in_text(text, model="gpt-3.5-turbo-0613"):
  """Returns the number of tokens in a string."""
  try:
      encoding = tiktoken.encoding_for_model(model)
  except KeyError:
      encoding = tiktoken.get_encoding("cl100k_base")
  return len(encoding.encode(text))


def concatenate_completions_for_reranking(completions):
    # concatenate completions and renumber
    # TODO: make 3 a parameter
    concatenated_completions = ''
    for i, completion in enumerate(completions):
        completion_content = completion['choices'][0]['message']['content']
        completion_content1, completion_content = completion_content.split('\n2.')
        completion_content1 = completion_content1.lstrip('1. ')
        completion_content2, completion_content = completion_content.split('\n3.')
        completion_content3 = completion_content
        
        concatenated_completions += f'{int_to_roman(i*3 + 1)}. {completion_content1}\n'
        concatenated_completions += f'{int_to_roman(i*3 + 2)}. {completion_content2}\n'
        concatenated_completions += f'{int_to_roman(i*3 + 3)}. {completion_content3}\n'
    return concatenated_completions, 3 * len(completions)

# Reranking utility functions
def int_to_roman(n: int) -> str:
    if not (0 < n < 4000):
        raise ValueError("Input integer must be between 1 and 3999")
    
    ints = (1000, 900,  500, 400, 100,  90, 50,  40, 10,  9,   5,  4,   1)
    nums = ('M',  'CM', 'D', 'CD', 'C', 'XC','L','XL','X','IX','V','IV','I')
    result = []
    
    for i in range(len(ints)):
        count = int(n / ints[i])
        result.append(nums[i] * count)
        n -= ints[i] * count
    
    return ''.join(result)

# parse reranked completions
def parse_reranked_completions(reranked_completion, n_completions):
    # First lstrip prelude before 1.
    # Find index of '1. ' and slice
    reranked_completion = reranked_completion[reranked_completion.index('1. '):]
    
    # Split on i. for i in range(1, n_completions + 1)
    # Then lstrip roman numeral and period
    roman_numeral_regex_pattern = r''
    for i in range(1, n_completions + 1):
        roman_numeral_regex_pattern += f'{int_to_roman(i)}\. '
        if i < n_completions:
            roman_numeral_regex_pattern += '|'
    roman_numeral_regex = re.compile(roman_numeral_regex_pattern)
    
    parsed_completions = []
    for i in range(1, n_completions + 1):
        if i == n_completions:
            parsed_completion = reranked_completion
        else:
            parsed_completion, reranked_completion = reranked_completion.split(f'{i+1}. ', maxsplit=1)
        if i == 1:
            parsed_completion = parsed_completion.lstrip('1. ')
        parsed_completion = re.sub(roman_numeral_regex, '', parsed_completion)
        parsed_completion = parsed_completion.lstrip(' "')
        parsed_completion = parsed_completion.rstrip(' \n"')
        parsed_completions.append(parsed_completion)
        
    return parsed_completions


def openai_prompt_select_passages_from_prior_art_portion(all_but_claims, claims, claim_element, prior_art_passage):
    # TODO: make 3 a parameter
    messages = [
        {"role": "system", "content": "You are an associate at a big law firm. Your task is to read the written description and of an invention disclosed in a patent and the full set of claims in the patent and interpret a single claim element that I select from the claims of the patent in light of the written description and the full set of claims. Then I'm going to give you a passage from of prior art and you need to pick three distinct sub-passages from the passage that you believe are the most semantically similar to the claim element interpreted in light of the written description and the full set of claims. Please output only those three passages, each preceded only by the numberings 1., 2., and 3. and a space."},
        {"role": "user", "content": f"Written description of invention: {all_but_claims}"},
        {"role": "user", "content": f"Full set of claims: {claims}"},
        {"role": "user", "content": f"Selected claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passage: {prior_art_passage}"},
    ]
    return messages

def openai_prompt_rank_selected_passages(all_but_claims, claims, claim_element, chosen_prior_art_passage, n_completions):
    messages = [
        {"role": "system", "content": f"Your task is to read the written description and of an invention disclosed in a patent and the full set of claims in the patent and interpret a single claim element that I select from the claims of the patent in light of the written description and the full set of claims. Then I'm going to give you {n_completions} sub-passages from the prior art passage i presented to you that you chose because they were the most semantically similar to the selected claim element interpreted in light of the written description and the full set of claims. Please rank these {n_completions} passages in descending order of similarity to the selected claim element interpreted in light of the written description and the full set of claims. I will give you passages numbered by roman numerals, for example: 'I. Passage One... II. Passage two... III. Passage three...'. Please output only the integer ranking following by the roman numeral and passage, for example: '1. II. Passage two ... 2. I. Passage one... 3. III Passage three...' if you believe the ranking should be passage two > passage one > passage three."},
        {"role": "user", "content": f"Written description of invention: {all_but_claims}"},
        {"role": "user", "content": f"Claim element: {claim_element}"},
        {"role": "user", "content": f"Prior art passages to rank: {chosen_prior_art_passage}"},
    ]
    return messages

def openai_chat_completion_request_with_retry(model_name, messages, backoff_factor=2, backoff_override=None):
    for i in range(1, 3):
        try:
            # completion = openai_sync_client().chat.completions.create(
            #     model=model_name,
            #     messages=messages
            # )
            completion = openai.ChatCompletion.create(
                model=model_name,
                messages=messages
            )
            return completion
        except openai.OpenAIError as e:
            if backoff_override is not None:
                time.sleep(backoff_override)
            else:
                time.sleep(backoff_factor**i)
            continue
        except Exception as e:
            raise e
    raise Exception("Failed to send chat completion request to OpenAI after 4 retries.")



async def aopenai_chat_completion_request_with_retry(model_name, messages, backoff_factor=2, backoff_override=None, n_retries=3, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None, **kwargs) -> dict:
    """Send a chat completion request to OpenAI with retries."""
    for i in range(1, 1 + n_retries):
        try:
            if rate_limiter is not None:
                num_tokens_in_messages = num_tokens_from_messages(messages)
                # NOTE: separating wait for tokens, the request, and removing tokens is prone to race conditions. We can tolerate occasionally making requests that receive rate limit errors so long as we retry them with backoff.
                if await rate_limiter.wait_for_tokens(num_tokens_in_messages):
                    # completion = await openai_async_client().chat.completions.create(
                    #     model=model_name,
                    #     messages=messages,
                    #     **kwargs
                    # )
                    completion = await openai.ChatCompletion.acreate(
                        model=model_name,
                        messages=messages,
                        **kwargs
                    )
                    n_tokens_in_completion = completion['usage']['completion_tokens']
                    await rate_limiter.remove_tokens(n_tokens_in_completion)
            else:
                # completion = await openai_async_client().chat.completions.create(
                #         model=model_name,
                #         messages=messages,
                #         **kwargs
                #     )
                completion = await openai.ChatCompletion.acreate(
                    model=model_name,
                    messages=messages,
                    **kwargs
                )
            return completion
        # except openai.OpenAIError as e:
        except openai.error.OpenAIError as e:
            logger.error(f"OpenAIError: {e}")
            if backoff_override is not None:
                jitter = random.uniform(-0.10, 0.10)
                await asyncio.sleep(backoff_override * (1 + jitter))
            else:
                jitter = random.uniform(-0.10, 0.10) 
                await asyncio.sleep(backoff_factor**i * (1 + jitter))
            continue
        except Exception as e:
            raise e
    raise Exception(f"Failed to send chat completion request to OpenAI after {n_retries} retries.")

def split_passage_to_meet_model_token_limit(passage: str, prompt_func: Callable[[str], list[dict]], model_name: str, min_n_splits: int = 1) -> list[str]:
    """ 
    Given a passage string, a prompt func that takes only the passage string, and a model, find the smallest number of splits necessary to get messages under token limit and return splits of the passage.
    """
    token_limit = model_name_to_token_limit[model_name]
    min_split_size = 256
    base_num_tokens = num_tokens_from_messages(
        prompt_func(' ' * min_split_size)
    )
    if base_num_tokens > token_limit:
        raise GenerationError("Prompt exceeds token limit by itself.")
    
    passage_splits = []
    passage_split_count = min_n_splits
    while True:
        # Check all splits
        candidate_splits = []
        for i in range(passage_split_count):
            if i == 0:
                candidate_split = passage[:len(passage) // passage_split_count]
            else:
                candidate_split = passage[len(passage) // passage_split_count * i:len(passage) // passage_split_count * (i+1)]
            candidate_splits.append(candidate_split)
        num_tokens = [num_tokens_from_messages(prompt_func(split)) for split in candidate_splits]
        if all([n <= token_limit - 0.05 * token_limit for n in num_tokens]):
            passage_splits = candidate_splits
            break
        passage_split_count += 1

    return passage_splits


def openai_model_id_from_completion(completion: dict) -> str:
    try:
        return f'{completion["model"]}-{completion["system_fingerprint"]}'
    except KeyError:
        return load_generator_config_from_env().model_name


async def agenerate_passages_for_prior_art_splits(prior_art_splits: list[str], prompt_func: Callable[[str], list[dict]], model_name: str, claim_element_id: int, prior_art_id: int, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None) -> list[GeneratedPassage | Exception]:
    # make parallel requests to openai api for prior art splits
    passages: list[GeneratedPassage | Exception] = []
    coros = []
    for prior_art_split in prior_art_splits:
        messages = prompt_func(prior_art_split)
        coros.append(aopenai_chat_completion_request_with_retry(model_name, messages, rate_limiter=rate_limiter))
        
    completions = await asyncio.gather(*coros, return_exceptions=True)
    for i, completion in enumerate(completions):
        if isinstance(completion, Exception):
            print(f'Failed to get completion for prior art split {i} of prior art {prior_art_id} for claim element {claim_element_id}')
            print(completion)
            passages.append(completion)
        else:
            selected_passages = completion['choices'][0]['message']['content']
            post_process_generated_passages_func = load_generator_config_from_env().post_process_generated_passages_func
            post_processed_passages = post_process_generated_passages_func(selected_passages)
            # if len(post_processed_passages) == 1:
            #     logger.error(f'Possible error post processing selected passages for claim element {claim_element_id} and prior art {prior_art_id}: \nCompletion: {selected_passages} \nPost-processed: {post_processed_passages}')

            for processed_passage in post_processed_passages:
                # We'll add the start line and end line later
                passages.append(
                    GeneratedPassage(
                        prior_art_source_id=prior_art_id,
                        text=processed_passage,
                        claim_element_id=claim_element_id,
                        model_id=openai_model_id_from_completion(completion)
                    )
                )

    passages = [p for p in passages if not isinstance(p, Exception)]


    return passages

async def agenerate_passages(claim_elements: list[tuple[int, str]], prior_art: tuple[int, str], prompt_func: Callable[[str], list[dict]], model_name: str, target_n_passages: int = 10, min_n_splits:int = 1, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None) -> AsyncIterator[tuple[int, tuple[int, list[str]]]]:
    tasks = []
    for i, claim_element in claim_elements:
        prior_art_splits = split_passage_to_meet_model_token_limit(prior_art[1], partial(prompt_func, claim_element), model_name, min_n_splits=min_n_splits)
        n_passages_per_split = max(target_n_passages // len(prior_art_splits), 1)
        tasks.append(asyncio.create_task(
            agenerate_passages_for_prior_art_splits(prior_art_splits, partial(prompt_func, claim_element, n_passages_per_split=n_passages_per_split), model_name, claim_element_id=i, prior_art_id=prior_art[0], rate_limiter=rate_limiter)
        ))
    
    for coro in asyncio.as_completed(tasks):
        yield await coro

def create_generate_passages_tasks(claim_elements: list[tuple[int, str]], prior_art: tuple[int, str], prompt_func: Callable[[str], list[dict]], model_name: str, target_n_passages: int = 10, min_n_splits:int = 1, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None) -> list[asyncio.Task]:
    tasks = []
    for i, claim_element in claim_elements:
        prior_art_splits = split_passage_to_meet_model_token_limit(prior_art[1], partial(prompt_func, claim_element), model_name, min_n_splits=min_n_splits)
        n_passages_per_split = max(target_n_passages // len(prior_art_splits), 1)
        tasks.append(asyncio.create_task(
            agenerate_passages_for_prior_art_splits(prior_art_splits, partial(prompt_func, claim_element, n_passages_per_split=n_passages_per_split), model_name, claim_element_id=i, prior_art_id=prior_art[0], rate_limiter=rate_limiter)
        ))

    return tasks


def summarize_patent_prompt_func(serialized_patent_spec: str, instruction: str = None):
    if instruction is None:
        instruction = "Summarize the following patent specification in approximately 250 words:"
    return [
        {"role": "user", "content": f"{instruction}"},
        {"role": "user", "content": serialized_patent_spec},
    ]


def summarize_summaries_prompt_func(summaries: list[str], instruction: str = None):
    if instruction is None:
        instruction = "Summarize the following partial summaries of sequential portions of a patent specification in approximately 250 words:"
    return [
        {"role": "user", "content": f"{instruction}"},
        {"role": "user", "content": '\n'.join(summaries)},
    ]

async def asummarize_patent(patent: parser.GoogleParsedPatent, rate_limiter: Optional[DistributedFixedWindowRateLimiter] = None):
    generator_config = load_generator_config_from_env()
    serialized_patent_spec = patent.specification

    split_instruction = "Summarize the following portion of a patent specification in approximately 250 words:"
    spec_splits = split_passage_to_meet_model_token_limit(serialized_patent_spec, partial(summarize_patent_prompt_func, instruction=split_instruction), generator_config.model_name)

    if len(spec_splits) == 1:
        summarization_completion = await aopenai_chat_completion_request_with_retry(
            generator_config.model_name, summarize_patent_prompt_func(serialized_patent_spec), rate_limiter=rate_limiter
        )
        summarization = summarization_completion['choices'][0]['message']['content']
    else:
        tasks = []
        # TODO: hack
        for spec_split in spec_splits[:10]:
            tasks.append(
                aopenai_chat_completion_request_with_retry(
                    generator_config.model_name, summarize_patent_prompt_func(spec_split, instruction=split_instruction), rate_limiter=rate_limiter
                )
            )
        completions = await asyncio.gather(*tasks)
        partial_summaries = []
        for completion in completions:
            partial_summaries.append(completion['choices'][0]['message']['content'])

        try:
            summarization_completion = await aopenai_chat_completion_request_with_retry(
                generator_config.model_name, summarize_summaries_prompt_func(partial_summaries), rate_limiter=rate_limiter
            )
            summarization = summarization_completion['choices'][0]['message']['content']
        except GenerationError as e:
            logger.error('Failed to summarize patent %s', patent.unique_id)
            summarization = ''

    return summarization

async def abulk_generate_passages(
        patent: tuple[int, parser.GoogleParsedPatent], 
        prior_art_sources: list[tuple[int, parser.GoogleParsedPatent]], claim_elements: list[tuple[int, str]],
        ranking_params: dict,
        patent_summary: Optional[str] = None,
        ) -> AsyncIterator[GeneratedPassage]:
    # TODO: this function should be refactored to separate concerns
    generator_config = load_generator_config_from_env()
    async with DistributedFixedWindowRateLimiter(token_rate_limit_by_model_name[generator_config.model_name], RATE_LIMIT_INTERVAL) as rate_limiter:
        if patent_summary is None:
            patent_summary = await asummarize_patent(patent[1])
        claims = parser.parse_claims_from_google_parsed_patent(patent[1])
        all_patent_claim_elements = []
        for claim in claims:
            all_patent_claim_elements.extend(claim.claim_elements)
        serialized_claims = '\n'.join(all_patent_claim_elements)
        tasks = []
        for prior_art_source_uid, prior_art_source in prior_art_sources:
            serialized_prior_art_source = prior_art_source.specification
            prompt_func = generator_config.generate_passages_prompt_func
            if prompt_func is prompt_summary_full_set_of_claims:
                prompt_func = partial(prompt_func, patent_summary, serialized_claims)
            else:
                raise NotImplementedError(f"Prompt func {prompt_func} not implemented.")

            tasks.extend(
                create_generate_passages_tasks(
                    claim_elements, (prior_art_source_uid, serialized_prior_art_source), 
                    prompt_func,
                    generator_config.model_name,
                    target_n_passages=generator_config.target_n_passages,
                    min_n_splits=generator_config.min_n_splits,
                    rate_limiter=rate_limiter
            ))
        
        for coro in asyncio.as_completed(tasks):
            generated_passages = await coro
            # TODO: once we yield the whole list of generated passages for each awaited task, we can do the ranking step in the caller
            # Find matching claim element
            claim_element_id = generated_passages[0].claim_element_id
            matching_claim_element = [c for c in claim_elements if c[0] == claim_element_id][0]
            if not all([p.claim_element_id == claim_element_id for p in generated_passages]):
                logger.error('Claim element ids do not match')
            
            logger.info(f'Ranking {len(generated_passages)} passages')
            ranked_generated_passages = await ranking.arank_passages(generated_passages, matching_claim_element[1], **ranking_params)
            for generated_passage in ranked_generated_passages:
                # TODO: this depends on parsing patent from pdf working properly
                # citation = None
                # try:
                #     citation = utils.cite_passage(generated_passage.text, patent[1])
                # except utils.CitationError as e:
                #     logger.error(e)
                # else:
                #     generated_passage.start_line = citation[0]
                #     generated_passage.end_line = citation[1]
                # TODO: this doesn't need to be a yield, should just return the whole passage set for each claim element
                yield generated_passage


import os
from logging import getLogger
import threading

import redis
import redis.asyncio as async_redis

from . import settings

logger = getLogger(__name__)

__all__ = ['r', 'AsyncRedisLock', 'async_redis']

logger.info(f"Connecting to redis at {settings.REDIS_HOST}")
r = redis.Redis(host=settings.REDIS_HOST, port=6379)

class AsyncRedisLock:
    def __init__(self, lock: async_redis.lock.Lock):
        self.lock = lock

    async def __aenter__(self):
        await self.lock.acquire()

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self.lock.release()